@@ -27,7 +27,7 @@ TRNS = (116, 82, 78, 83)
 ACTL_CHUNK = (0, 0, 0, 8, 97, 99, 84, 76, 0, 0, 0, 2, 0, 0, 0, 0, 243, 141, 147, 112)
 FCTL = (102, 99, 84, 76)
 IDAT = (73, 68, 65, 84)
-FDAT = (102, 100, 65, 84, 0, 0, 0, 2)
+FDAT = (102, 100, 65, 84)
 IEND_CHUNK = (0, 0, 0, 0, 73, 69, 78, 68, 174, 66, 96, 130)
 CRC_MASK = 4294967295
 
@@ -79,10 +79,14 @@ class PngWriter:
         self.actl_chunk = bytearray(ACTL_CHUNK)
         self.idat = bytearray(IDAT)
         self.fdat = bytearray(FDAT)
+        self.fdat2 = bytearray(FDAT + (0, 0, 0, 2))
         self.iend_chunk = bytearray(IEND_CHUNK)
 
-    def write_image(self, udg_array, img_file, scale, x, y, width, height, full_size, palette, attr_map, trans, flash_rect):
-        bit_depth, plte_chunk, frame1, frame2, frame2_rect = self._build_image_data(udg_array, scale, x, y, width, height, full_size, palette, attr_map, trans, flash_rect)
+    def write_image(self, frames, img_file, palette, attr_map, trans, flash_rect):
+        bit_depth, palette_size = self._get_bit_depth(palette)
+        frame1 = frames[0]
+        width, height = frame1.width, frame1.height
+        frame1_data, frame2_data, frame2_rect = self._build_image_data(frame1, palette_size, bit_depth, attr_map, flash_rect)
 
         # PNG signature
         img_file.write(self.png_signature)
@@ -91,25 +95,39 @@ class PngWriter:
         self._write_ihdr_chunk(img_file, width, height, bit_depth)
 
         # PLTE
-        self._write_chunk(img_file, plte_chunk)
+        self._write_plte_chunk(img_file, palette)
 
         # tRNS
         if trans and self.alpha != 255:
             self._write_chunk(img_file, self.trns + [self.alpha])
 
-        # acTL and fcTL
-        if frame2:
+        # acTL
+        if len(frames) == 1 and flash_rect:
             img_file.write(self.actl_chunk)
-            self._write_fctl_chunk(img_file, 0, width, height)
+        elif len(frames) > 1:
+            actl_chunk = (97, 99, 84, 76, 0, 0, 0, len(frames), 0, 0, 0, 0)
+            self._write_chunk(img_file, actl_chunk)
+
+        # fcTL
+        if len(frames) > 1 or flash_rect:
+            seq_num = 0
+            self._write_fctl_chunk(img_file, seq_num, frame1.delay, width, height)
 
         # IDAT
-        self._write_img_data_chunk(img_file, self.idat + frame1)
+        self._write_img_data_chunk(img_file, self.idat + frame1_data)
 
         # fcTL and fdAT
-        if frame2:
-            f2_x, f2_y, f2_w, f2_h = frame2_rect
-            self._write_fctl_chunk(img_file, 1, f2_w, f2_h, f2_x, f2_y)
-            self._write_img_data_chunk(img_file, self.fdat + frame2)
+        if len(frames) == 1 and flash_rect:
+            f2_x_offset, f2_y_offset, f2_width, f2_height = frame2_rect
+            self._write_fctl_chunk(img_file, 1, frame1.delay, f2_width, f2_height, f2_x_offset, f2_y_offset)
+            self._write_img_data_chunk(img_file, self.fdat2 + frame2_data)
+        for frame in frames[1:]:
+            frame_data = self._build_image_data(frame, palette_size, bit_depth, attr_map)[0]
+            seq_num += 1
+            self._write_fctl_chunk(img_file, seq_num, frame.delay, frame.width, frame.height)
+            seq_num += 1
+            fdat = self.fdat + bytearray(self._to_bytes(seq_num))
+            self._write_img_data_chunk(img_file, fdat + frame_data)
 
         # IEND
         img_file.write(self.iend_chunk)
@@ -238,20 +256,25 @@ class PngWriter:
         data.extend((0, 0, 0)) # compression, filter and interlace methods
         self._write_chunk(img_file, data)
 
-    def _write_fctl_chunk(self, img_file, frame_num, width, height, x_offset=0, y_offset=0):
+    def _write_plte_chunk(self, img_file, palette):
+        data = list(PLTE)
+        data.extend(palette)
+        self._write_chunk(img_file, data)
+
+    def _write_fctl_chunk(self, img_file, seq_num, delay, width, height, x_offset=0, y_offset=0):
         data = list(FCTL) # chunk type
-        data.extend(self._to_bytes(frame_num)) # frame number
+        data.extend(self._to_bytes(seq_num)) # sequence number
         data.extend(self._to_bytes(width)) # width
         data.extend(self._to_bytes(height)) # height
         data.extend(self._to_bytes(x_offset)) # x offset
         data.extend(self._to_bytes(y_offset)) # y offset
-        data.extend((0, 8)) # delay numerator
-        data.extend((0, 25)) # delay denominator
+        data.extend((delay // 256, delay % 256)) # delay numerator
+        data.extend((0, 100)) # delay denominator
         data.append(0) # frame disposal operation
         data.append(0) # frame blending operation
         self._write_chunk(img_file, data)
 
-    def _build_image_data(self, udg_array, scale, x, y, width, height, full_size, palette, attr_map, trans, flash_rect):
+    def _get_bit_depth(self, palette):
         palette_size = len(palette) // 3
         if palette_size > 4:
             bit_depth = 4
@@ -259,8 +282,14 @@ class PngWriter:
             bit_depth = 2
         else:
             bit_depth = 1
-        plte_chunk = list(PLTE)
-        plte_chunk.extend(palette)
+        return bit_depth, palette_size
+
+    def _build_image_data(self, frame, palette_size, bit_depth, attr_map, flash_rect=None):
+        udg_array = frame.udgs
+        scale = frame.scale
+        x, y, width, height = frame.x, frame.y, frame.width, frame.height
+        trans = frame.trans
+        full_size = not frame.cropped
 
         frame2 = frame2_rect = None
         if flash_rect:
@@ -296,7 +325,7 @@ class PngWriter:
                 f2_build_method = f2_method(len(f2_udg_array[0]) * len(f2_udg_array), len(attr_map), scale)
             frame2 = f2_build_method(f2_udg_array, scale, attr_map, trans, True, x + f2_x, y + f2_y, f2_w, f2_h, bit_depth)
 
-        return bit_depth, plte_chunk, frame1, frame2, frame2_rect
+        return frame1, frame2, frame2_rect
 
     def _get_crc(self, byte_list):
         crc = CRC_MASK
