@@ -16,16 +16,16 @@
 # You should have received a copy of the GNU General Public License along with
 # SkoolKit. If not, see <http://www.gnu.org/licenses/>.
 
+import sys
 import os
 from os.path import isfile, isdir, basename, dirname
-import posixpath
 import shutil
 import time
 import argparse
 
-from . import PACKAGE_DIR, VERSION, show_package_dir, write, write_line, get_class, SkoolKitError
+from . import PACKAGE_DIR, VERSION, show_package_dir, write, write_line, get_class, normpath, SkoolKitError
 from .image import ImageWriter
-from .skoolhtml import join, FileInfo
+from .skoolhtml import FileInfo
 from .skoolparser import SkoolParser, CASE_UPPER, CASE_LOWER, BASE_10, BASE_16
 from .refparser import RefParser
 
@@ -39,7 +39,27 @@ SEARCH_DIRS = (
     '/usr/share/skoolkit',
     os.path.join(PACKAGE_DIR, 'resources')
 )
-CONFIG = 'Config'
+
+SEARCH_DIRS_MSG = """
+skool2html.py searches the following directories for skool files, ref files,
+CSS files, JavaScript files, font files, and files listed in the [Resources]
+section of the ref file:
+
+""".lstrip()
+
+def show_search_dirs():
+    write(SEARCH_DIRS_MSG)
+    prefix = '- '
+    write_line(prefix + 'The directory that contains the skool or ref file named on the command line')
+    for search_dir in SEARCH_DIRS:
+        if not search_dir:
+            search_dir = 'The current working directory'
+        elif not os.path.split(search_dir)[0]:
+            search_dir = os.path.join('.', search_dir)
+        else:
+            search_dir = os.path.normpath(search_dir)
+        write_line(prefix + search_dir)
+    sys.exit(0)
 
 def notify(notice):
     if verbose:
@@ -58,14 +78,12 @@ def clock(operation, prefix, *args, **kwargs):
     return result
 
 def find(fname, search_dir=None):
-    if fname is None:
-        return
     if search_dir:
         search_dirs = [search_dir]
     else:
         search_dirs = []
     search_dirs.extend(SEARCH_DIRS)
-    for f in [join(d, fname) for d in search_dirs]:
+    for f in [os.path.join(d, fname) for d in search_dirs]:
         if isfile(f):
             return f
 
@@ -77,10 +95,6 @@ def add_lines(ref_parser, config_specs, section=None):
         if section is None or section_name == section:
             ref_parser.add_line(section_name, line)
 
-def get_config(ref_parser, config_specs):
-    add_lines(ref_parser, config_specs, CONFIG)
-    return ref_parser.get_dictionary(CONFIG)
-
 def get_colours(colour_specs):
     colours = {}
     for k, v in colour_specs.items():
@@ -102,90 +116,102 @@ def get_colours(colour_specs):
             raise SkoolKitError("Invalid colour spec: {}={}".format(k, v))
     return colours
 
-def copy_resources(search_dir, root_dir, fnames, dest, themes=None, suffix=None, indent=0):
-    if fnames:
-        actual_files = []
-        for fname in fnames.split(';'):
-            files = []
-            f = find(fname, search_dir)
-            if f:
-                files.append(f)
-            else:
-                raise SkoolKitError('{0}: file not found'.format(fname))
-            if themes and fname.endswith(suffix):
-                for theme in themes:
-                    f = find('{0}-{1}{2}'.format(fname[:-len(suffix)], theme, suffix), search_dir)
-                    if f:
-                        files.append(f)
+def copy_resource(fname, root_dir, dest_dir, indent=0):
+    base_f = basename(fname)
+    dest_f = normpath(root_dir, dest_dir, base_f)
+    if not isfile(dest_f) or os.stat(fname).st_mtime > os.stat(dest_f).st_mtime:
+        fname_n = normpath(fname)
+        dest_d = dirname(dest_f)
+        if isfile(dest_d):
+            raise SkoolKitError("Cannot copy {0} to {1}: {1} is not a directory".format(fname_n, normpath(dest_dir)))
+        elif not isdir(dest_d):
+            os.makedirs(dest_d)
+        notify('{}Copying {} to {}'.format(indent * ' ', fname_n, dest_f))
+        shutil.copy2(fname, dest_f)
+
+def copy_resources(search_dir, root_dir, fnames, dest_dir, themes=None, suffix=None, single_css=None, indent=0):
+    if not fnames:
+        return
+    files = []
+    actual_files = []
+    for fname in fnames.split(';'):
+        f = find(fname, search_dir)
+        if f:
+            files.append(f)
+        else:
+            raise SkoolKitError('{}: file not found'.format(normpath(fname)))
+        if themes and fname.endswith(suffix):
+            for theme in themes:
+                f = find('{0}-{1}{2}'.format(fname[:-len(suffix)], theme, suffix), search_dir)
+                if f:
+                    files.append(f)
+        if not single_css:
             for f in files:
-                base_f = basename(f)
-                actual_files.append(base_f)
-                html_f = posixpath.normpath(join(root_dir, dest, base_f))
-                if not isfile(html_f) or os.stat(f).st_mtime > os.stat(html_f).st_mtime:
-                    notify('{}Copying {} to {}'.format(indent * ' ', f, html_f))
-                    shutil.copy2(f, html_f)
-        return ';'.join(actual_files)
+                copy_resource(f, root_dir, dest_dir, indent)
+                actual_files.append(basename(f))
+            files = []
+    if single_css:
+        dest_css = normpath(root_dir, dest_dir, single_css)
+        if isdir(dest_css):
+            raise SkoolKitError("Cannot write CSS file '{}': {} already exists and is a directory".format(normpath(single_css), dest_css))
+        with open(dest_css, 'w') as css:
+            for f in files:
+                notify('{}Appending {} to {}'.format(indent * ' ', normpath(f), dest_css))
+                with open(f) as src:
+                    for line in src:
+                        css.write(line)
+                css.write('\n')
+        return single_css
+    return ';'.join(actual_files)
 
 def get_prefix(fname):
     if '.' in fname:
         return fname.rsplit('.', 1)[0]
     return fname
 
-def process_file(infile, topdir, files, case, base, pages, config_specs, new_images, css_theme, create_labels, asm_labels):
-    reffile = skoolfile = None
+def process_file(infile, topdir, files, case, base, pages, config_specs, new_images, css_theme, create_labels, asm_labels, single_css):
+    skoolfile_f = reffile_f = None
+    ref_search_dir = ''
     if infile.endswith('.ref'):
-        reffile = infile
-    else:
-        skoolfile = infile
-
-    ref_parser = RefParser()
-    reffile_f = find(reffile)
-    config = {}
-    search_dir = None
-
-    prefix = None
-    if reffile:
-        if reffile_f is None:
-            raise SkoolKitError('{0}: file not found'.format(reffile))
-        search_dir = dirname(reffile_f)
-        ref_parser.parse(reffile_f)
-        config = get_config(ref_parser, config_specs)
-        prefix = get_prefix(basename(reffile))
-        skoolfile = config.get('SkoolFile', '{0}.skool'.format(prefix))
-
-    if skoolfile == '-':
-        skoolfile_f = skoolfile
-    else:
-        skoolfile_f = find(skoolfile, search_dir)
-    if skoolfile_f is None:
-        raise SkoolKitError('{0}: file not found'.format(skoolfile))
-
-    if prefix is None:
-        if skoolfile_f == '-':
-            prefix = 'program'
-        else:
-            prefix = get_prefix(basename(skoolfile_f))
-
-    if reffile is None and skoolfile != '-':
-        search_dir = dirname(skoolfile_f)
-        reffile_f = find('{0}.ref'.format(prefix), search_dir)
+        reffile_f = find(infile)
         if reffile_f:
-            ref_parser.parse(reffile_f)
-            config = get_config(ref_parser, config_specs)
-            search_dir = dirname(reffile_f)
+            ref_search_dir = dirname(reffile_f)
+            prefix = get_prefix(basename(reffile_f))
+    elif infile == '-':
+        skoolfile_f = infile
+        prefix = 'program'
+    else:
+        skoolfile_f = find(infile)
+        if skoolfile_f:
+            ref_search_dir = dirname(skoolfile_f)
+            prefix = get_prefix(basename(skoolfile_f))
+            reffile_f = find('{}.ref'.format(prefix), ref_search_dir)
+            if reffile_f:
+                ref_search_dir = dirname(reffile_f)
+    if skoolfile_f is reffile_f is None:
+        raise SkoolKitError('{}: file not found'.format(normpath(infile)))
 
-    reffiles = [posixpath.join(search_dir, f) for f in os.listdir(search_dir or '.') if f.startswith(prefix) and f.endswith('.ref')]
-    reffiles = [f for f in reffiles if isfile(f)]
-    if reffile_f in reffiles:
-        reffiles.remove(reffile_f)
-    reffiles.sort()
+    reffiles = []
+    base_ref = prefix + '.ref'
+    for f in sorted(os.listdir(ref_search_dir or '.')):
+        if isfile(os.path.join(ref_search_dir, f)) and f.endswith('.ref') and f.startswith(prefix) and f != base_ref:
+            reffiles.append(normpath(ref_search_dir, f))
+    if reffile_f:
+        reffiles.insert(0, normpath(reffile_f))
+    ref_parser = RefParser()
     for oreffile_f in reffiles:
         ref_parser.parse(oreffile_f)
-    if reffile_f:
-        reffiles.insert(0, reffile_f)
     add_lines(ref_parser, config_specs)
 
-    notify('Using skool file: {0}'.format(skoolfile_f))
+    config = ref_parser.get_dictionary('Config')
+    if skoolfile_f is None:
+        skoolfile = config.get('SkoolFile', '{}.skool'.format(prefix))
+        skoolfile_f = find(skoolfile, ref_search_dir)
+        if skoolfile_f is None:
+            raise SkoolKitError('{}: file not found'.format(normpath(skoolfile)))
+
+    skoolfile_n = normpath(skoolfile_f)
+    notify('Using skool file: {}'.format(skoolfile_n))
     if reffiles:
         if len(reffiles) > 1:
             suffix = 's'
@@ -193,7 +219,7 @@ def process_file(infile, topdir, files, case, base, pages, config_specs, new_ima
             suffix = ''
         notify('Using ref file{0}: {1}'.format(suffix, ', '.join(reffiles)))
     else:
-        notify('Found no ref file for {0}'.format(skoolfile_f))
+        notify('Found no ref file for {}'.format(skoolfile_n))
 
     html_writer_class = get_class(config.get('HtmlWriterClass', 'skoolkit.skoolhtml.HtmlWriter'))
     game_dir = config.get('GameDir', prefix)
@@ -217,9 +243,9 @@ def process_file(infile, topdir, files, case, base, pages, config_specs, new_ima
             raise SkoolKitError('Invalid page ID: {0}'.format(page_id))
     pages = pages or all_page_ids
 
-    write_disassembly(html_writer, files, search_dir, pages, css_theme)
+    write_disassembly(html_writer, files, ref_search_dir, pages, css_theme, single_css)
 
-def write_disassembly(html_writer, files, search_dir, pages, css_theme):
+def write_disassembly(html_writer, files, search_dir, pages, css_theme, single_css):
     game_dir = html_writer.file_info.game_dir
     paths = html_writer.paths
     game_vars = html_writer.game_vars
@@ -231,12 +257,20 @@ def write_disassembly(html_writer, files, search_dir, pages, css_theme):
         os.makedirs(odir)
 
     # Copy CSS and font files if necessary
-    html_writer.set_style_sheet(copy_resources(search_dir, odir, game_vars.get('StyleSheet'), paths.get('StyleSheetPath', ''), css_theme, '.css'))
+    html_writer.set_style_sheet(copy_resources(search_dir, odir, game_vars.get('StyleSheet'), paths.get('StyleSheetPath', ''), css_theme, '.css', single_css))
     copy_resources(search_dir, odir, game_vars.get('Font'), paths.get('FontPath', ''))
 
+    # Copy resources named in the [Resources] section
+    resources = html_writer.ref_parser.get_dictionary('Resources')
+    for f, dest_dir in resources.items():
+        fname = find(f, search_dir)
+        if not fname:
+            raise SkoolKitError('Cannot copy resource "{}": file not found'.format(normpath(f)))
+        copy_resource(fname, odir, dest_dir)
+
     # Write disassembly files
     if 'd' in files:
-        clock(html_writer.write_asm_entries, '  Writing disassembly files in {0}'.format(join(game_dir, html_writer.code_path)))
+        clock(html_writer.write_asm_entries, '  Writing disassembly files in {}'.format(normpath(game_dir, html_writer.code_path)))
 
     # Write the memory map files
     if 'm' in files:
@@ -244,32 +278,32 @@ def write_disassembly(html_writer, files, search_dir, pages, css_theme):
             map_details = html_writer.memory_maps[map_name]
             if html_writer.should_write_map(map_details):
                 map_path = html_writer.get_map_path(map_details)
-                clock(html_writer.write_map, '  Writing {0}'.format(join(game_dir, map_path)), map_details)
+                clock(html_writer.write_map, '  Writing {}'.format(normpath(game_dir, map_path)), map_details)
 
     # Write pages defined in the ref file
     if 'P' in files:
         for page_id in pages:
             page_details = html_writer.pages[page_id]
             copy_resources(search_dir, odir, page_details.get('JavaScript'), paths.get('JavaScriptPath', ''), indent=2)
-            clock(html_writer.write_page, '  Writing {0}'.format(join(game_dir, paths[page_id])), page_id)
+            clock(html_writer.write_page, '  Writing {}'.format(normpath(game_dir, paths[page_id])), page_id)
 
     # Write other files
     if 'G' in files and html_writer.has_gbuffer():
-        clock(html_writer.write_gbuffer, '  Writing {0}'.format(join(game_dir, paths['GameStatusBuffer'])))
+        clock(html_writer.write_gbuffer, '  Writing {}'.format(normpath(game_dir, paths['GameStatusBuffer'])))
     if 'g' in files and html_writer.graphics:
-        clock(html_writer.write_graphics, '  Writing {0}'.format(join(game_dir, paths['Graphics'])))
+        clock(html_writer.write_graphics, '  Writing {}'.format(normpath(game_dir, paths['Graphics'])))
     if 'B' in files and html_writer.graphic_glitches:
-        clock(html_writer.write_graphic_glitches, '  Writing {0}'.format(join(game_dir, paths['GraphicGlitches'])))
+        clock(html_writer.write_graphic_glitches, '  Writing {}'.format(normpath(game_dir, paths['GraphicGlitches'])))
     if 'c' in files and html_writer.changelog:
-        clock(html_writer.write_changelog, '  Writing {0}'.format(join(game_dir, paths['Changelog'])))
+        clock(html_writer.write_changelog, '  Writing {}'.format(normpath(game_dir, paths['Changelog'])))
     if 'b' in files and html_writer.bugs:
-        clock(html_writer.write_bugs, '  Writing {0}'.format(join(game_dir, paths['Bugs'])))
+        clock(html_writer.write_bugs, '  Writing {}'.format(normpath(game_dir, paths['Bugs'])))
     if 't' in files and html_writer.facts:
-        clock(html_writer.write_facts, '  Writing {0}'.format(join(game_dir, paths['Facts'])))
+        clock(html_writer.write_facts, '  Writing {}'.format(normpath(game_dir, paths['Facts'])))
     if 'y' in files and html_writer.glossary:
-        clock(html_writer.write_glossary, '  Writing {0}'.format(join(game_dir, paths['Glossary'])))
+        clock(html_writer.write_glossary, '  Writing {}'.format(normpath(game_dir, paths['Glossary'])))
     if 'p' in files and html_writer.pokes:
-        clock(html_writer.write_pokes, '  Writing {0}'.format(join(game_dir, paths['Pokes'])))
+        clock(html_writer.write_pokes, '  Writing {}'.format(normpath(game_dir, paths['Pokes'])))
     if 'o' in files:
         mode = html_writer.parser.mode
         for code_id, code in html_writer.other_code:
@@ -283,28 +317,25 @@ def write_disassembly(html_writer, files, search_dir, pages, css_theme):
                 'AsmPath': asm_path,
                 'Title': code['Title']
             }
-            clock(html_writer2.write_map, '    Writing {0}'.format(join(game_dir, map_path)), map_details)
-            clock(html_writer2.write_entries, '    Writing disassembly files in {0}'.format(join(game_dir, asm_path)), asm_path, map_path, code['Header'])
+            clock(html_writer2.write_map, '    Writing {}'.format(normpath(game_dir, map_path)), map_details)
+            clock(html_writer2.write_entries, '    Writing disassembly files in {}'.format(normpath(game_dir, asm_path)), asm_path, map_path, code['Header'])
 
     # Write index.html
     if 'i' in files:
-        clock(html_writer.write_index, '  Writing {0}'.format(join(game_dir, paths['GameIndex'])))
+        clock(html_writer.write_index, '  Writing {}'.format(normpath(game_dir, paths['GameIndex'])))
 
 def run(files, options):
     if options.output_dir in (None, '.'):
         topdir = ''
     else:
-        pdir, bdir = os.path.split(options.output_dir)
-        if not bdir:
-            pdir, bdir = os.path.split(pdir)
-        topdir = join(pdir, bdir)
+        topdir = normpath(options.output_dir)
         if not isdir(topdir):
             # Create the top-level directory
             notify('Creating directory {0}'.format(topdir))
             os.makedirs(topdir)
 
     for infile in files:
-        process_file(infile, topdir, options.files, options.case, options.base, options.pages, options.config_specs, options.new_images, options.themes, options.create_labels, options.asm_labels)
+        process_file(infile, topdir, options.files, options.case, options.base, options.pages, options.config_specs, options.new_images, options.themes, options.create_labels, options.asm_labels, options.single_css)
 
 def main(args):
     global verbose, show_timings
@@ -331,6 +362,8 @@ def main(args):
                        help="Write the disassembly in decimal")
     group.add_argument('-H', '--hex', dest='base', action='store_const', const=BASE_16,
                        help="Write the disassembly in hexadecimal")
+    group.add_argument('-j', '--join-css', dest='single_css', metavar='NAME',
+                       help="Concatenate CSS files into a single file with this name")
     group.add_argument('-l', '--lower', dest='case', action='store_const', const=CASE_LOWER,
                        help="Write the disassembly in lower case")
     group.add_argument('-o', '--rebuild-images', dest='new_images', action='store_true',
@@ -344,6 +377,8 @@ def main(args):
                             "file(s)")
     group.add_argument('-q', '--quiet', dest='verbose', action='store_false',
                        help="Be quiet")
+    group.add_argument('-s', '--search-dirs', dest='search_dirs', action='store_true',
+                       help="Show the locations skool2html.py searches for resources")
     group.add_argument('-t', '--time', dest='show_timings', action='store_true',
                        help="Show timings")
     group.add_argument('-T', '--theme', dest='themes', metavar='THEME', action='append',
@@ -367,6 +402,8 @@ def main(args):
     namespace, unknown_args = parser.parse_known_args(args)
     if namespace.package_dir:
         show_package_dir()
+    if namespace.search_dirs:
+        show_search_dirs()
     if unknown_args or not namespace.infiles:
         parser.exit(2, parser.format_help())
     verbose, show_timings = namespace.verbose, namespace.show_timings
