@@ -19,9 +19,10 @@
 import re
 import inspect
 
-from . import warn, write_text, wrap, parse_int, get_chr, SkoolParsingError
-from .skoolmacro import parse_ints, parse_params, get_params, MacroParsingError, UnsupportedMacroError
-from .skoolparser import TableParser, ListParser, TABLE_MARKER, TABLE_END_MARKER, LIST_MARKER, LIST_END_MARKER, HTML_MACRO_DELIMITERS
+from . import warn, write_text, wrap, get_chr, SkoolParsingError
+from . import skoolmacro
+from .skoolmacro import MacroParsingError, UnsupportedMacroError
+from .skoolparser import TableParser, ListParser, TABLE_MARKER, TABLE_END_MARKER, LIST_MARKER, LIST_END_MARKER
 
 UDGTABLE_MARKER = '#UDGTABLE'
 
@@ -145,16 +146,6 @@ class AsmWriter:
     def write_line(self, s):
         write_text('{0}{1}'.format(s, self.end))
 
-    def _expand_item_macro(self, text, index, default):
-        end, params, p_text = parse_params(text, index, default)
-        return end, p_text
-
-    def _expand_unsupported_macro(self, text, index):
-        if self.handle_unsupported_macros:
-            end, params, p_text = parse_params(text, index, chars=',:;-x{}')
-            return end, ''
-        raise UnsupportedMacroError()
-
     def pop_snapshot(self):
         """Discard the current memory snapshot and replace it with the one that
         was most recently saved (by
@@ -180,140 +171,68 @@ class AsmWriter:
                 self.macros[macro] = method
 
     def expand_bug(self, text, index):
-        # #BUG[#name][(link text)]
-        return self._expand_item_macro(text, index, 'bug')
+        end, item, link_text = skoolmacro.parse_bug(text, index)
+        return end, link_text or 'bug'
 
     def expand_call(self, text, index):
-        # #CALL:methodName(args)
-        macro = '#CALL'
-        if text[index] != ':':
-            raise MacroParsingError("Malformed macro: {0}{1}...".format(macro, text[index]))
-        end, method_name, arg_string = parse_params(text, index + 1, chars='_')
-        if not hasattr(self, method_name):
-            self.warn("Unknown method name in {0} macro: {1}".format(macro, method_name))
-            return end, ''
-        method = getattr(self, method_name)
-        if not inspect.ismethod(method):
-            raise MacroParsingError("Uncallable method name: {0}".format(method_name))
-        if arg_string is None:
-            raise MacroParsingError("No argument list specified: {0}{1}".format(macro, text[index:end]))
-        args = get_params(arg_string)
-        retval = method(*args)
-        if retval is None:
-            retval = ''
-        return end, retval
+        return skoolmacro.parse_call(text, index, self)
 
     def expand_chr(self, text, index):
-        # #CHRnum or #CHR(num)
-        if text[index:].startswith('('):
-            offset = 1
-        else:
-            offset = 0
-        end, num = parse_ints(text, index + offset, 1)
-        return end + offset, get_chr(num)
+        end, num = skoolmacro.parse_chr(text, index)
+        return end, get_chr(num)
 
     def expand_d(self, text, index):
-        # #Daddr
-        end, addr = parse_ints(text, index, 1)
-        entry = self.parser.get_entry(addr)
-        if entry:
-            if entry.description:
-                return end, entry.description
-            raise MacroParsingError('Entry at {0} has no description'.format(addr))
-        raise MacroParsingError('Cannot determine description for nonexistent entry at {0}'.format(addr))
+        return skoolmacro.parse_d(text, index, self.parser)
 
     def expand_erefs(self, text, index):
-        # #EREFSaddr
-        end, address = parse_ints(text, index, 1)
-        ereferrers = self.parser.get_entry_point_refs(address)
-        if text[index] == '$':
-            addr_fmt = '${0:04X}'
-        else:
-            addr_fmt = '{0}'
-        if not ereferrers:
-            raise MacroParsingError('Entry point at {0} has no referrers'.format(addr_fmt.format(address)))
-        ereferrers.sort()
-        rep = 'routine at '
-        if len(ereferrers) > 1:
-            rep = 'routines at '
-            rep += ', '.join('#R{0}'.format(addr_fmt.format(addr)) for addr in ereferrers[:-1])
-            rep += ' and '
-        addr = ereferrers[-1]
-        rep += '#R{0}'.format(addr_fmt.format(addr))
-        return end, rep
+        return skoolmacro.parse_erefs(text, index, self.parser)
 
     def expand_fact(self, text, index):
-        # #FACT[#name][(link text)]
-        return self._expand_item_macro(text, index, 'fact')
+        end, item, link_text = skoolmacro.parse_fact(text, index)
+        return end, link_text or 'fact'
 
     def expand_font(self, text, index):
-        return self._expand_unsupported_macro(text, index)
+        # #FONT[:(text)]addr[,chars,attr,scale][{X,Y,W,H}][(fname)]
+        if self.handle_unsupported_macros:
+            if index < len(text) and text[index] == ':':
+                index, message = skoolmacro.get_text_param(text, index + 1)
+            end, params, p_text = skoolmacro.parse_params(text, index, chars=',{}')
+            return end, ''
+        raise UnsupportedMacroError()
 
     def expand_html(self, text, index):
-        # #HTML(text)
-        delim1 = text[index]
-        delim2 = HTML_MACRO_DELIMITERS.get(delim1, delim1)
-        start = index + 1
-        end = text.find(delim2, start)
-        if end < start:
-            raise MacroParsingError("No terminating delimiter: #HTML{0}".format(text[index:]))
-        return end + 1, ''
+        end, message = skoolmacro.parse_html(text, index)
+        return end, ''
 
     def expand_link(self, text, index):
-        # #LINK:PageId[#name](link text)
-        macro = '#LINK'
-        if text[index] != ':':
-            raise MacroParsingError("Malformed macro: {0}{1}...".format(macro, text[index]))
-        end, params, link_text = parse_params(text, index + 1)
+        end, page_id, anchor, link_text = skoolmacro.parse_link(text, index)
         if not link_text:
-            raise MacroParsingError("No link text specified: {0}{1}".format(macro, text[index:end]))
+            raise MacroParsingError("Blank link text: #LINK{}".format(text[index:end]))
         return end, link_text
 
     def expand_poke(self, text, index):
-        # #POKE[#name][(link text)]
-        return self._expand_item_macro(text, index, 'poke')
+        end, item, link_text = skoolmacro.parse_poke(text, index)
+        return end, link_text or 'poke'
 
     def expand_pokes(self, text, index):
-        # #POKESaddr,byte[,length,step][;addr,byte[,length,step];...]
-        end, addr, byte, length, step = parse_ints(text, index, 4, (1, 1))
-        self.snapshot[addr:addr + length * step:step] = [byte] * length
-        while end < len(text) and text[end] == ';':
-            end, addr, byte, length, step = parse_ints(text, end + 1, 4, (1, 1))
-            self.snapshot[addr:addr + length * step:step] = [byte] * length
-        return end, ''
+        return skoolmacro.parse_pokes(text, index, self.snapshot)
 
     def expand_pops(self, text, index):
-        # #POPS
-        self.pop_snapshot()
-        return index, ''
+        return skoolmacro.parse_pops(text, index, self)
 
     def expand_pushs(self, text, index):
-        # #PUSHS[name]
-        end, name, p_text = parse_params(text, index)
-        self.push_snapshot(name)
-        return end, ''
+        return skoolmacro.parse_pushs(text, index, self)
 
     def expand_r(self, text, index):
-        # #Raddr[@code][#anchor][(link text)]
-        end, params, p_text = parse_params(text, index, chars='@')
-        if p_text:
-            return end, p_text
-        anchor_index = params.find('#')
-        if anchor_index > 0:
-            params = params[:anchor_index]
-        code_id = ''
-        code_id_index = params.find('@')
-        if code_id_index > 0:
-            code_id = params[code_id_index + 1:]
-            params = params[:code_id_index]
-        addr_str = params
-        address = parse_int(addr_str)
+        end, addr_str, address, code_id, anchor, link_text = skoolmacro.parse_r(text, index)
+        if link_text:
+            return end, link_text
         if code_id:
             return end, self.parser.get_instruction_addr_str(address, code_id)
         label = self.labels.get(address)
         if label is None:
             if self.base_address <= address <= self.end_address:
-                self.warn('Could not convert address {0} to label'.format(addr_str))
+                self.warn('Could not convert address {} to label'.format(addr_str))
             label = self.parser.get_instruction_addr_str(address)
             if label is None:
                 if addr_str.startswith('$'):
@@ -323,60 +242,39 @@ class AsmWriter:
         return end, label
 
     def expand_refs(self, text, index):
-        # #REFSaddr[(prefix)]
-        end, addr_str, prefix = parse_params(text, index, '')
-        address = parse_int(addr_str)
-        entry = self.entries.get(address)
-        if not entry:
-            raise MacroParsingError('No entry at {0}'.format(addr_str))
-        if text[index] == '$':
-            addr_fmt = '${0:04X}'
-        else:
-            addr_fmt = '{0}'
-        referrers = [ref.address for ref in entry.referrers]
-        if referrers:
-            referrers.sort()
-            rep = '{0} routine at '.format(prefix).lstrip()
-            if len(referrers) > 1:
-                rep = '{0} routines at '.format(prefix).lstrip()
-                rep += ', '.join('#R{0}'.format(addr_fmt.format(addr)) for addr in referrers[:-1])
-                rep += ' and '
-            addr = referrers[-1]
-            rep += '#R{0}'.format(addr_fmt.format(addr))
-        else:
-            rep = 'Not used directly by any other routines'
-        return end, rep
+        return skoolmacro.parse_refs(text, index, self.entries)
 
     def expand_reg(self, text, index):
-        # #REGreg
-        end, reg, p_text = parse_params(text, index, chars="'")
-        if not reg:
-            raise MacroParsingError('Missing register argument')
-        if len(reg) > 3 or any([char not in "abcdefhlirspxy'" for char in reg]):
-            raise MacroParsingError('Bad register: "{0}"'.format(reg))
-        if self.lower:
-            reg_name = reg.lower()
-        else:
-            reg_name = reg.upper()
-        return end, reg_name
+        return skoolmacro.parse_reg(text, index, self.lower)
 
     def expand_scr(self, text, index):
-        return self._expand_unsupported_macro(text, index)
+        # #SCR[scale,x,y,w,h,dfAddr,afAddr][{X,Y,W,H}][(fname)]
+        if self.handle_unsupported_macros:
+            end, params, p_text = skoolmacro.parse_params(text, index, chars=',{}')
+            return end, ''
+        raise UnsupportedMacroError()
 
     def expand_space(self, text, index):
-        # #SPACE[num] or #SPACE([num])
-        if text[index:].startswith('('):
-            offset = 1
-        else:
-            offset = 0
-        end, num_sp = parse_ints(text, index + offset, 1, (1,))
-        return end + offset, ''.ljust(num_sp)
+        end, num_sp = skoolmacro.parse_space(text, index)
+        return end, ' ' * num_sp
 
     def expand_udg(self, text, index):
-        return self._expand_unsupported_macro(text, index)
+        # #UDGaddr[,attr,scale,step,inc,flip,rotate][:maskAddr[,maskStep]][{X,Y,W,H}][(fname)]
+        if self.handle_unsupported_macros:
+            end, params, p_text = skoolmacro.parse_params(text, index, chars=',:{}')
+            return end, ''
+        raise UnsupportedMacroError()
 
     def expand_udgarray(self, text, index):
-        return self._expand_unsupported_macro(text, index)
+        # #UDGARRAYwidth[,attr,scale,step,inc,flip,rotate];addr1[,attr1,step1,inc1][:maskAddr1[,maskStep1]];...[{X,Y,W,H}](fname)
+        # #UDGARRAY*frame1[,delay1];frame2[,delay2];...(fname)
+        if self.handle_unsupported_macros:
+            if index < len(text) and text[index] == '*':
+                end, params, p_text = skoolmacro.parse_params(text, index, except_chars=' (')
+            else:
+                end, params, p_text = skoolmacro.parse_params(text, index, chars=',:;-x{}')
+            return end, ''
+        raise UnsupportedMacroError()
 
     def expand(self, text):
         if text.find('#') < 0:
@@ -405,7 +303,10 @@ class AsmWriter:
         index = 0
         while text.find(marker, index) >= 0:
             block_index = text.index(marker, index)
-            block_end_index = text.index(end_marker, block_index) + len(end_marker)
+            try:
+                block_end_index = text.index(end_marker, block_index) + len(end_marker)
+            except ValueError:
+                raise SkoolParsingError("Missing end marker: {}...".format(text[block_index:block_index + len(marker) + 15]))
             block_indexes.append(block_index)
             block_indexes.append(block_end_index)
             index = block_end_index
@@ -437,10 +338,11 @@ class AsmWriter:
         for block in self.extract_blocks(text):
             if block.startswith(TABLE_MARKER):
                 table_lines = self.table_writer.format_table(self.expand(block[len(TABLE_MARKER):]))
-                table_width = max([len(line) for line in table_lines])
-                if table_width > width:
-                    self.warn('Table in entry at {0} is {1} characters wide'.format(self.entry.address, table_width))
-                lines.extend(table_lines)
+                if table_lines:
+                    table_width = max([len(line) for line in table_lines])
+                    if table_width > width:
+                        self.warn('Table in entry at {0} is {1} characters wide'.format(self.entry.address, table_width))
+                    lines.extend(table_lines)
             elif block.startswith(LIST_MARKER):
                 list_obj = self.list_parser.parse_list(self.expand(block[len(LIST_MARKER):]))
                 for item in list_obj.items:
@@ -591,10 +493,12 @@ class TableWriter:
         return rendered
 
     def _create_table_text(self):
+        lines = []
         max_row_index = len(self.table.rows)
+        if max_row_index == 0:
+            return lines
         separator_row_indexes = set((0, max_row_index))
         separator_row_indexes.update([i + 1 for i in self.table.get_header_rows()])
-        lines = []
         for row_index in range(max_row_index):
             if row_index in separator_row_indexes:
                 lines.append(self._create_row_separator(row_index))
