@@ -74,12 +74,12 @@ class CtlWriter:
 
     def _write_entry_asm_directive(self, entry, directive, value=None):
         if self.write_asm_dirs:
-            self._write_asm_directive(directive, self.addr_str(entry.address), value)
+            self._write_asm_directive(directive, self.addr_str(entry.address, False), value)
 
     def _write_instruction_asm_directives(self, instruction):
         if not self.write_asm_dirs:
             return
-        address = self.addr_str(instruction.address)
+        address = self.addr_str(instruction.address, False)
         for directive, value in instruction.asm_directives:
             self._write_asm_directive(directive, address, value)
 
@@ -151,15 +151,20 @@ class CtlWriter:
                             length = int(sub_blocks[j + offset][1][0].address) - int(instructions[0].address)
                         elif k + 1 < len(sections):
                             length = int(sections[k + 1][1][0].address) - int(instructions[0].address)
-                        comment = ''
-                        if instructions[0].comment and COMMENTS in self.elements:
-                            comment = instructions[0].comment.text
-                        if self.need_sub_block(ctl, entry_ctl, comment, j, instructions):
-                            self.write_sub_block(ctl, entry_ctl, comment, instructions, length)
+                        comment_text = ''
+                        comment = instructions[0].comment
+                        if comment and COMMENTS in self.elements:
+                            comment_text = comment.text
+                            if comment.rowspan > 1 and not comment_text.replace('.', ''):
+                                comment_text = '.' + comment_text
+                        if self.need_sub_block(ctl, entry_ctl, comment_text, j, instructions):
+                            self.write_sub_block(ctl, entry_ctl, comment_text, instructions, length)
 
-    def addr_str(self, address):
+    def addr_str(self, address, pad=True):
         if self.write_hex:
-            return '${0:04X}'.format(int(address))
+            return '${:04X}'.format(address)
+        if pad:
+            return '{:05d}'.format(address)
         return str(address)
 
     def need_sub_block(self, ctl, entry_ctl, comment, index, instructions):
