@@ -26,9 +26,10 @@ import os.path
 from os.path import isfile, isdir, basename
 import inspect
 
-from . import VERSION, warn, parse_int, SkoolKitError, SkoolParsingError
-from .skoolmacro import parse_ints, parse_params, get_params, MacroParsingError, UnsupportedMacroError
-from .skoolparser import TableParser, ListParser, CASE_LOWER, HTML_MACRO_DELIMITERS
+from . import VERSION, warn, get_int_param, parse_int, SkoolKitError, SkoolParsingError
+from . import skoolmacro
+from .skoolmacro import MacroParsingError, UnsupportedMacroError
+from .skoolparser import TableParser, ListParser, CASE_LOWER
 
 #: The ID of the main disassembly.
 MAIN_CODE_ID = 'main'
@@ -104,6 +105,7 @@ class HtmlWriter:
         self.file_info = file_info
         self.image_writer = image_writer
         self.default_image_format = image_writer.default_format
+        self.frames = {}
         self.case = case
         self.base = skool_parser.base
 
@@ -198,6 +200,9 @@ class HtmlWriter:
         """
         return
 
+    def warn(self, s):
+        warn(s)
+
     def clone(self, skool_parser, code_id):
         the_clone = self.__class__(skool_parser, self.ref_parser, self.file_info, self.image_writer, self.case, code_id)
         the_clone.set_style_sheet(self.game_vars['StyleSheet'])
@@ -504,8 +509,12 @@ class HtmlWriter:
                     rotated[0].append(udgs[j][i])
             udgs[:] = rotated
 
-    def get_font_udg_array(self, address, chars, attr):
-        return 8 * chars, [[Udg(attr, self.snapshot[a:a + 8]) for a in range(address, address + 8 * chars, 8)]]
+    def get_font_udg_array(self, address, attr, message):
+        udgs = []
+        for c in message:
+            a = address + 8 * (ord(c) - 32)
+            udgs.append(Udg(attr, self.snapshot[a:a + 8]))
+        return [udgs]
 
     def write_asm_entries(self):
         self.write_entries(self.code_path, self.paths[P_MEMORY_MAP])
@@ -1021,6 +1030,14 @@ class HtmlWriter:
         ofile.write('</tr>\n')
         ofile.write('</table>\n')
 
+    def _get_image_format(self, image_path):
+        img_file_ext = image_path.lower()[-4:]
+        if img_file_ext == '.png':
+            return 'png'
+        if img_file_ext == '.gif':
+            return 'gif'
+        raise SkoolKitError('Unsupported image file format: {}'.format(image_path))
+
     def write_image(self, image_path, udgs, crop_rect=(), scale=2, mask=False):
         """Create an image and write it to a file.
 
@@ -1039,15 +1056,23 @@ class HtmlWriter:
         :param mask: Whether to apply masks to the tiles in the image.
         """
         if self.image_writer:
-            img_file_ext = image_path.lower()[-4:]
-            if img_file_ext == '.png':
-                img_format = 'png'
-            elif img_file_ext == '.gif':
-                img_format = 'gif'
-            else:
-                raise SkoolKitError('Unsupported image file format: {0}'.format(image_path))
+            frame = Frame(udgs, scale, mask, *crop_rect)
+            self.write_animated_image(image_path, [frame])
+
+    def write_animated_image(self, image_path, frames):
+        """Create an animated image and write it to a file.
+
+        :param image_path: The full path of the file to which to write the
+                           image (relative to the root directory of the
+                           disassembly).
+        :param frames: A list of the frames (instances of
+                       :class:`~skoolkit.skoolhtml.Frame`) from which to build
+                       the image.
+        """
+        if self.image_writer:
+            img_format = self._get_image_format(image_path)
             f = self.file_info.open_file(image_path, mode='wb')
-            self.image_writer.write_image(udgs, f, img_format, scale, mask, *crop_rect)
+            self.image_writer.write_image(frames, f, img_format)
             f.close()
 
     def _create_macros(self):
@@ -1108,29 +1133,35 @@ class HtmlWriter:
         html.append('</ul>')
         return '\n'.join(html)
 
-    def _get_udg_addresses(self, addr, width):
-        if type(addr) is int:
-            return [addr]
+    def _get_udg_addresses(self, addr_spec, width):
+        if type(addr_spec) is int:
+            return [addr_spec]
         num = 1
+        addr = addr_spec
         if 'x' in addr:
-            addr, num = addr.split('x')
-            num = parse_int(num)
-        if '-' in addr:
-            elements = [parse_int(n) for n in addr.split('-')]
-            if len(elements) < 3:
-                elements.append(1)
-            if len(elements) < 4:
-                elements.append(elements[2] * width)
-            address, end_address, h_step, v_step = elements
-            addresses = []
-            while address <= end_address:
-                addresses.append(address)
-                if len(addresses) % width:
-                    address += h_step
-                else:
-                    address += v_step - (width - 1) * h_step
-        else:
-            addresses = [parse_int(addr)]
+            addr, num = addr_spec.split('x', 1)
+            try:
+                num = get_int_param(num)
+            except ValueError:
+                raise MacroParsingError("Invalid address range specification: {}".format(addr_spec))
+        try:
+            elements = [get_int_param(n) for n in addr.split('-', 3)]
+        except ValueError:
+            raise MacroParsingError("Invalid address range specification: {}".format(addr_spec))
+        if len(elements) < 2:
+            elements.append(elements[0])
+        if len(elements) < 3:
+            elements.append(1)
+        if len(elements) < 4:
+            elements.append(elements[2] * width)
+        address, end_address, h_step, v_step = elements
+        addresses = []
+        while address <= end_address:
+            addresses.append(address)
+            if len(addresses) % width:
+                address += h_step
+            else:
+                address += v_step - (width - 1) * h_step
         return addresses * num
 
     def img_element(self, cwd, image_path, alt=None):
@@ -1171,11 +1202,12 @@ class HtmlWriter:
                 return join(self.paths[path_id], '{0}{1}'.format(fname, suffix))
             raise SkoolKitError("Unknown path ID '{0}' for image file '{1}'".format(path_id, fname))
 
-    def parse_image_params(self, text, index, num, defaults=(), path_id=DEF_IMG_PATH, fname='', chars=''):
+    def parse_image_params(self, text, index, num, defaults=(), path_id=DEF_IMG_PATH, fname='', chars='', ints=None):
         """Parse a string of the form ``params[{X,Y,W,H}][(fname)]``. The
         parameter string ``params`` may contain comma-separated values and will
-        be parsed until either the end of the text is reached, or an invalid
-        character is encountered.
+        be parsed until either the end is reached, or an invalid character is
+        encountered. The default set of valid characters consists of the comma,
+        '$', the digits 0-9, and the letters A-F and a-f.
 
         :param text: The text to parse.
         :param index: The index at which to start parsing.
@@ -1184,162 +1216,105 @@ class HtmlWriter:
         :param path_id: The ID of the target directory for the image file (as
                         defined in the :ref:`paths` section of the `ref` file).
         :param fname: The default base name of the image file.
-        :param chars: Characters to consider valid in the parameter string in
-                      addition to the comma, '$', the digits 0-9, and the
-                      letters A-F and a-f.
+        :param chars: Characters to consider valid in addition to those in the
+                      default set.
+        :param ints: A list of the indexes (0-based) of the parameters that
+                     must evaluate to an integer; if `None`, every parameter
+                     must evaluate to an integer.
         :return: A list of the form
-                 ``[end, image_path, crop_rect, value1, value2...]``, where
-                 ``end`` is the index at which parsing terminated,
-                 ``image_path`` is the full path of the image file (relative to
-                 the root directory of the disassembly), ``crop_rect`` is
-                 ``(X, Y, W, H)``, and ``value1``, ``value2`` etc. are the
-                 parameter values.
+                 ``[end, image_path, crop_rect, value1, value2...]``, where:
+
+                 * ``end`` is the index at which parsing terminated
+                 * ``image_path`` is either the full path of the image file
+                   (relative to the root directory of the disassembly) or
+                   ``fname`` (if `path_id` is blank or `None`)
+                 * ``crop_rect`` is ``(X, Y, W, H)``
+                 * ``value1``, ``value2`` etc. are the parameter values
         """
         valid_chars = '$0123456789abcdefABCDEF,' + chars
-        param_string = ''
-        commas = 0
+        end, param_string, p_text = skoolmacro.parse_params(text, index, only_chars=valid_chars)
+        params = skoolmacro.get_params(param_string, num, defaults, ints)
+        if len(params) > num:
+            raise MacroParsingError("Too many parameters (expected {}): '{}'".format(num, text[index:end]))
 
-        while index < len(text) and text[index] in valid_chars:
-            char = text[index]
-            if char == ',':
-                commas += 1
-                if commas == num:
-                    break
-            param_string += char
-            index += 1
-        params = get_params(param_string, num, defaults)
-
-        x = y = 0
-        width = height = None
-        if index < len(text) and text[index] == '{':
-            index, x, y, width, height = parse_ints(text, index + 1, 4, (0, 0, None, None))
-            index += 1
-
-        if index < len(text) and text[index] == '(':
-            end = text.index(')', index) + 1
-            fname = text[index + 1:end - 1]
+        if end < len(text) and text[end] == '{':
+            end, x, y, width, height = skoolmacro.parse_ints(text, end + 1, 4, (0, 0, None, None))
+            end, param_string, p_text = skoolmacro.parse_params(text, end, only_chars='}')
         else:
-            end = index
-        img_path = self.image_path(fname, path_id)
+            x = y = 0
+            width = height = None
+
+        if p_text:
+            fname = p_text
+        if path_id:
+            img_path = self.image_path(fname, path_id)
+        else:
+            img_path = fname
 
         return [end, img_path, (x, y, width, height)] + params
 
-    def _expand_item_macro(self, macro, text, index, cwd, items, path_id, def_link_text):
-        end, anchor, link_text = parse_params(text, index)
-        if anchor and link_text == '':
+    def _expand_item_macro(self, item, link_text, cwd, items, path_id):
+        if item and link_text == '':
             for name, title, contents in items:
-                if anchor[1:] == name:
+                if item == name:
                     link_text = title
                     break
             else:
-                raise MacroParsingError("Cannot determine title of item: {0}{1}()".format(macro, anchor))
-        if link_text is None:
-            link_text = def_link_text
+                raise MacroParsingError("Cannot determine title of item '{}'".format(item))
+        if item:
+            anchor = '#' + item
+        else:
+            anchor = ''
         items_file = FileInfo.relpath(cwd, self.paths[path_id])
-        link = '<a class="link" href="{0}{1}">{2}</a>'.format(items_file, anchor, link_text)
-        return end, link
+        return '<a class="link" href="{}{}">{}</a>'.format(items_file, anchor, link_text)
 
     def expand_bug(self, text, index, cwd):
-        # #BUG[#name][(link text)]
-        return self._expand_item_macro('#BUG', text, index, cwd, self.bugs, P_BUGS, 'bug')
+        end, item, link_text = skoolmacro.parse_bug(text, index)
+        return end, self._expand_item_macro(item, link_text, cwd, self.bugs, P_BUGS)
 
     def expand_call(self, text, index, cwd):
-        # #CALL:methodName(args)
-        macro = '#CALL'
-        if text[index] != ':':
-            raise MacroParsingError("Malformed macro: {0}{1}...".format(macro, text[index]))
-        end, method_name, arg_string = parse_params(text, index + 1, chars='_')
-        if not hasattr(self, method_name):
-            warn("Unknown method name in {0} macro: {1}".format(macro, method_name))
-            return end, ''
-        method = getattr(self, method_name)
-        if not inspect.ismethod(method):
-            raise MacroParsingError("Uncallable method name: {0}".format(method_name))
-        if arg_string is None:
-            raise MacroParsingError("No argument list specified: {0}{1}".format(macro, text[index:end]))
-        args = [cwd] + get_params(arg_string)
-        retval = method(*args)
-        if retval is None:
-            retval = ''
-        return end, retval
+        return skoolmacro.parse_call(text, index, self, cwd)
 
     def expand_chr(self, text, index, cwd):
-        # #CHRnum or #CHR(num)
-        if text[index:].startswith('('):
-            offset = 1
-        else:
-            offset = 0
-        end, num = parse_ints(text, index + offset, 1)
-        return end + offset, '&#{0};'.format(num)
+        end, num = skoolmacro.parse_chr(text, index)
+        return end, '&#{};'.format(num)
 
     def expand_d(self, text, index, cwd):
-        # #Daddr
-        end, addr = parse_ints(text, index, 1)
-        entry = self.get_entry(addr)
-        if entry:
-            if entry.description:
-                return end, entry.description
-            raise MacroParsingError('Entry at {0} has no description'.format(addr))
-        raise MacroParsingError('Cannot determine description for nonexistent entry at {0}'.format(addr))
+        return skoolmacro.parse_d(text, index, self)
 
     def expand_erefs(self, text, index, cwd):
-        # #EREFSaddr
-        end, address = parse_ints(text, index, 1)
-        ereferrers = self.get_entry_point_refs(address)
-        if not ereferrers:
-            raise MacroParsingError('Entry point at {0} has no referrers'.format(address))
-        if len(ereferrers) == 1:
-            html = 'routine at '
-        else:
-            ereferrers.sort()
-            html = 'routines at '
-            html += ', '.join('#R{}'.format(addr) for addr in ereferrers[:-1])
-            html += ' and '
-        html += '#R{}'.format(ereferrers[-1])
-        return end, html
+        return skoolmacro.parse_erefs(text, index, self)
 
     def expand_fact(self, text, index, cwd):
-        # #FACT[#name][(link text)]
-        return self._expand_item_macro('#FACT', text, index, cwd, self.facts, P_FACTS, 'fact')
+        end, item, link_text = skoolmacro.parse_fact(text, index)
+        return end, self._expand_item_macro(item, link_text, cwd, self.facts, P_FACTS)
 
     def expand_font(self, text, index, cwd):
-        # #FONTaddr,chars[,attr,scale][{X,Y,W,H}][(fname)]
-        end, img_path, crop_rect, addr, chars, attr, scale = self.parse_image_params(text, index, 4, (56, 2), 'FontImagePath', 'font')
+        # #FONT[:(text)]addr[,chars,attr,scale][{X,Y,W,H}][(fname)]
+        if index < len(text) and text[index] == ':':
+            index, message = skoolmacro.get_text_param(text, index + 1)
+            if not message:
+                raise MacroParsingError("Empty message: {}".format(text[index - 2:index]))
+        else:
+            message = ''.join([chr(n) for n in range(32, 128)])
+
+        end, img_path, crop_rect, addr, chars, attr, scale = self.parse_image_params(text, index, 4, (len(message), 56, 2), 'FontImagePath', 'font')
         if self.need_image(img_path):
-            full_width, font_udg_array = self.get_font_udg_array(addr, chars, attr)
-            if crop_rect[2] is None:
-                crop_rect = (crop_rect[0], crop_rect[1], full_width * scale, crop_rect[3])
-            self.write_image(img_path, font_udg_array, crop_rect, scale)
+            udg_array = self.get_font_udg_array(addr, attr, message[:chars])
+            self.write_image(img_path, udg_array, crop_rect, scale)
         return end, self.img_element(cwd, img_path)
 
     def expand_html(self, text, index, cwd):
-        # #HTML(text)
-        delim1 = text[index]
-        delim2 = HTML_MACRO_DELIMITERS.get(delim1, delim1)
-        start = index + 1
-        end = text.find(delim2, start)
-        if end < start:
-            raise MacroParsingError("No terminating delimiter: {0}{1}".format('#HTML', text[index:]))
-        return end + 1, text[start:end]
+        return skoolmacro.parse_html(text, index)
 
     def expand_link(self, text, index, cwd):
-        # #LINK:PageId[#name](link text)
-        macro = '#LINK'
-        if text[index] != ':':
-            raise MacroParsingError("Malformed macro: {0}{1}...".format(macro, text[index]))
-        end, page_id, link_text = parse_params(text, index + 1)
-        anchor = ''
-        if '#' in page_id:
-            page_id, anchor = page_id.split('#')
-            anchor = '#' + anchor
+        end, page_id, anchor, link_text = skoolmacro.parse_link(text, index)
         if page_id not in self.paths:
-            raise MacroParsingError("Unknown page ID: {0}".format(page_id))
+            raise MacroParsingError("Unknown page ID: {}".format(page_id))
         if link_text == '':
             link_text = self.links[page_id][0]
         href = FileInfo.relpath(cwd, self.paths[page_id])
-        if not link_text:
-            raise MacroParsingError("No link text: {0}{1}".format(macro, text[index:end]))
-        link = '<a class="link" href="{0}{1}">{2}</a>'.format(href, anchor, link_text)
+        link = '<a class="link" href="{}{}">{}</a>'.format(href, anchor, link_text)
         return end, link
 
     def expand_list(self, text, index, cwd):
@@ -1348,101 +1323,47 @@ class HtmlWriter:
         return end, self.build_list(list_obj)
 
     def expand_poke(self, text, index, cwd):
-        # #POKE[#name][(link text)]
-        return self._expand_item_macro('#POKE', text, index, cwd, self.pokes, P_POKES, 'poke')
+        end, item, link_text = skoolmacro.parse_poke(text, index)
+        return end, self._expand_item_macro(item, link_text, cwd, self.pokes, P_POKES)
 
     def expand_pokes(self, text, index, cwd):
-        # #POKESaddr,byte[,length,step][;addr,byte[,length,step];...]
-        end, addr, byte, length, step = parse_ints(text, index, 4, (1, 1))
-        self.snapshot[addr:addr + length * step:step] = [byte] * length
-        while end < len(text) and text[end] == ';':
-            end, addr, byte, length, step = parse_ints(text, end + 1, 4, (1, 1))
-            self.snapshot[addr:addr + length * step:step] = [byte] * length
-        return end, ''
+        return skoolmacro.parse_pokes(text, index, self.snapshot)
 
     def expand_pops(self, text, index, cwd):
-        # #POPS
-        self.pop_snapshot()
-        return index, ''
+        return skoolmacro.parse_pops(text, index, self)
 
     def expand_pushs(self, text, index, cwd):
-        # #PUSHS[name]
-        end, name, p_text = parse_params(text, index)
-        self.push_snapshot(name)
-        return end, ''
+        return skoolmacro.parse_pushs(text, index, self)
 
     def expand_r(self, text, index, cwd):
-        # #Raddr[@code][#anchor][(link text)]
-        end, params, link_text = parse_params(text, index, chars='@')
-        anchor = ''
-        anchor_index = params.find('#')
-        if anchor_index > 0:
-            anchor = params[anchor_index:]
-            params = params[:anchor_index]
-        code_id = ''
-        code_id_index = params.find('@')
-        if code_id_index > 0:
-            code_id = params[code_id_index + 1:]
-            params = params[:code_id_index]
+        end, addr_str, address, code_id, anchor, link_text = skoolmacro.parse_r(text, index)
         if code_id:
             code_path = self.get_code_path(code_id)
             if code_path is None:
                 raise MacroParsingError("Could not find code path for '{}' disassembly".format(code_id))
         else:
             code_path = self.code_path
-        addr_str = params
-        address = parse_int(addr_str)
         container = self.parser.get_container(address, code_id)
         if not code_id and not container:
-            raise MacroParsingError('Could not find routine file containing {0}'.format(addr_str))
+            raise MacroParsingError('Could not find routine file containing {}'.format(addr_str))
         inst_addr_str = self.parser.get_instruction_addr_str(address, code_id)
         if container:
             container_address = container.address
         else:
             container_address = address
         if address != container_address:
-            anchor = '#{0}'.format(address)
+            anchor = '#{}'.format(address)
         asm_label = self.parser.get_asm_label(address)
         ref_file = FileInfo.asm_relpath(cwd, container_address, code_path)
-        link = '<a class="link" href="{0}{1}">{2}</a>'.format(ref_file, anchor, link_text or asm_label or inst_addr_str)
+        link = '<a class="link" href="{}{}">{}</a>'.format(ref_file, anchor, link_text or asm_label or inst_addr_str)
         return end, link
 
     def expand_refs(self, text, index, cwd):
-        # #REFSaddr[(prefix)]
-        end, addr_str, prefix = parse_params(text, index, '')
-        address = parse_int(addr_str)
-        entry = self.entries.get(address)
-        if not entry:
-            raise MacroParsingError('No entry at {0}'.format(addr_str))
-        referrers = [ref.address for ref in entry.referrers]
-        if referrers:
-            if len(referrers) == 1:
-                html = ('{0} routine at '.format(prefix)).lstrip()
-            else:
-                referrers.sort()
-                html = ('{0} routines at '.format(prefix)).lstrip()
-                html += ', '.join('#R{}'.format(addr) for addr in referrers[:-1])
-                html += ' and '
-            html += '#R{}'.format(referrers[-1])
-        else:
-            html = 'Not used directly by any other routines'
-        return end, html
+        return skoolmacro.parse_refs(text, index, self.entries)
 
     def expand_reg(self, text, index, cwd):
-        # #REGreg
-        end = index
-        while end < len(text) and text[end] in "abcdefhlirspxy'":
-            end += 1
-        reg = text[index:end]
-        if not reg:
-            raise MacroParsingError('Missing register argument')
-        if len(reg) > 3:
-            raise MacroParsingError('Bad register: "{0}"'.format(reg))
-        if self.case == CASE_LOWER:
-            reg_name = reg.lower()
-        else:
-            reg_name = reg.upper()
-        return end, '<span class="register">{0}</span>'.format(reg_name)
+        end, reg = skoolmacro.parse_reg(text, index, self.case == CASE_LOWER)
+        return end, '<span class="register">{}</span>'.format(reg)
 
     def expand_scr(self, text, index, cwd):
         # #SCR[scale,x,y,w,h,dfAddr,afAddr][{X,Y,W,H}][(fname)]
@@ -1452,13 +1373,8 @@ class HtmlWriter:
         return end, self.img_element(cwd, scr_path)
 
     def expand_space(self, text, index, cwd):
-        # #SPACE[num] or #SPACE([num])
-        if text[index:].startswith('('):
-            offset = 1
-        else:
-            offset = 0
-        end, num_sp = parse_ints(text, index + offset, 1, (1,))
-        return end + offset, '&#160;' * num_sp
+        end, num_sp = skoolmacro.parse_space(text, index)
+        return end, '&#160;' * num_sp
 
     def expand_table(self, text, index, cwd):
         # #TABLE[(class[,col1class[,col2class...]])]<rows>TABLE#
@@ -1492,18 +1408,52 @@ class HtmlWriter:
             self.write_image(udg_path, [[udg]], crop_rect, scale, mask)
         return end, self.img_element(cwd, udg_path)
 
+    def _expand_udgarray_with_frames(self, text, index, cwd):
+        end, frame_params, fname = skoolmacro.parse_params(text, index, except_chars=' (')
+        if not fname:
+            raise MacroParsingError('Missing filename: #UDGARRAY{}'.format(text[index:end]))
+        img_path = self.image_path(fname, 'UDGImagePath')
+        if self.need_image(img_path):
+            frames = []
+            default_delay = 32 # 0.32s
+            for frame_param in frame_params[1:].split(';'):
+                elements = frame_param.rsplit(',', 1)
+                if len(elements) == 2:
+                    frame_id = elements[0]
+                    delay = parse_int(elements[1])
+                    if delay is None:
+                        raise MacroParsingError('Invalid delay parameter: "{}"'.format(elements[1]))
+                    default_delay = delay
+                else:
+                    frame_id = frame_param
+                    delay = default_delay
+                if not frame_id:
+                    raise MacroParsingError('Missing frame ID: #UDGARRAY{}'.format(text[index:end]))
+                try:
+                    frame = self.frames[frame_id]
+                except KeyError:
+                    raise MacroParsingError('No such frame: "{}"'.format(frame_id))
+                frame.delay = delay
+                frames.append(frame)
+            self.write_animated_image(img_path, frames)
+        return end, self.img_element(cwd, img_path)
+
     def expand_udgarray(self, text, index, cwd):
         # #UDGARRAYwidth[,attr,scale,step,inc,flip,rotate];addr1[,attr1,step1,inc1][:maskAddr1[,maskStep1]];...[{X,Y,W,H}](fname)
-        udg_path_id = 'UDGImagePath'
-        end, img_path, crop_rect, width, def_attr, scale, def_step, def_inc, flip, rotate = self.parse_image_params(text, index, 7, (56, 2, 1, 0, 0, 0), udg_path_id)
+        # #UDGARRAY*frame1[,delay1];frame2[,delay2];...(fname)
+        if index < len(text) and text[index] == '*':
+            return self._expand_udgarray_with_frames(text, index, cwd)
+
+        udg_path_id = None
+        end, fname, crop_rect, width, def_attr, scale, def_step, def_inc, flip, rotate = self.parse_image_params(text, index, 7, (56, 2, 1, 0, 0, 0), udg_path_id)
         udg_array = [[]]
         has_masks = False
         while end < len(text) and text[end] == ';':
-            end, img_path, crop_rect, address, attr, step, inc = self.parse_image_params(text, end + 1, 4, (def_attr, def_step, def_inc), udg_path_id, chars='-x')
+            end, fname, crop_rect, address, attr, step, inc = self.parse_image_params(text, end + 1, 4, (def_attr, def_step, def_inc), udg_path_id, chars='-x', ints=(1, 2, 3))
             udg_addresses = self._get_udg_addresses(address, width)
             mask_addresses = []
             if end < len(text) and text[end] == ':':
-                end, img_path, crop_rect, mask_addr, mask_step = self.parse_image_params(text, end + 1, 2, (step,), udg_path_id, chars='-x')
+                end, fname, crop_rect, mask_addr, mask_step = self.parse_image_params(text, end + 1, 2, (step,), udg_path_id, chars='-x', ints=(1,))
                 mask_addresses = self._get_udg_addresses(mask_addr, width)
             has_masks = len(mask_addresses) > 0
             mask_addresses += [None] * (len(udg_addresses) - len(mask_addresses))
@@ -1516,15 +1466,30 @@ class HtmlWriter:
                     udg_array.append([udg])
                 else:
                     udg_array[-1].append(udg)
-        if img_path is None:
+
+        if not fname:
             raise MacroParsingError('Missing filename: #UDGARRAY{0}'.format(text[index:end]))
-        if self.need_image(img_path):
+
+        fname, sep, frame = fname.partition('*')
+        if sep and not frame:
+            frame = fname
+        if not fname and not frame:
+            raise MacroParsingError('Missing filename or frame ID: #UDGARRAY{}'.format(text[index:end]))
+
+        img_path = self.image_path(fname, 'UDGImagePath')
+        need_image = img_path and self.need_image(img_path)
+        if frame or need_image:
             if flip:
                 self.flip_udgs(udg_array, flip)
             if rotate:
                 self.rotate_udgs(udg_array, rotate)
-            self.write_image(img_path, udg_array, crop_rect, scale, has_masks)
-        return end, self.img_element(cwd, img_path)
+            if frame:
+                self.frames[frame] = Frame(udg_array, scale, has_masks, *crop_rect)
+            if need_image:
+                self.write_image(img_path, udg_array, crop_rect, scale, has_masks)
+        if img_path:
+            return end, self.img_element(cwd, img_path)
+        return end, ''
 
     def expand_udgtable(self, text, index, cwd):
         return self.expand_table(text, index, cwd)
@@ -1659,3 +1624,61 @@ class Udg(object):
         if self.mask:
             return Udg(self.attr, self.data[:], self.mask[:])
         return Udg(self.attr, self.data[:])
+
+class Frame(object):
+    """Create a frame of an animated image.
+
+    :param udgs: The two-dimensional array of tiles (instances of
+                 :class:`~skoolkit.skoolhtml.Udg`) from which to build the
+                 frame.
+    :param scale: The scale of the frame.
+    :param mask: Whether to apply masks to the tiles in the frame.
+    :param x: The x-coordinate of the top-left pixel to include in the frame.
+    :param y: The y-coordinate of the top-left pixel to include in the frame.
+    :param width: The width of the frame; if `None`, the maximum width
+                  (derived from `x` and width of the array of tiles) is used.
+    :param height: The height of the frame; if `None`, the maximum height
+                   (derived from `y` and height of the array of tiles) is used.
+    :param delay: The delay between this frame and the next in 1/100ths of a
+                  second.
+    """
+    def __init__(self, udgs, scale=1, mask=False, x=0, y=0, width=None, height=None, delay=32):
+        self._udgs = udgs
+        self._scale = scale
+        self.mask = mask
+        self._x = x
+        self._y = y
+        self._full_width = 8 * len(udgs[0]) * scale
+        self._full_height = 8 * len(udgs) * scale
+        self._width = min(width or self._full_width, self._full_width - x)
+        self._height = min(height or self._full_height, self._full_height - y)
+        self.delay = delay
+        self.trans = 0
+
+    @property
+    def udgs(self):
+        return self._udgs
+
+    @property
+    def scale(self):
+        return self._scale
+
+    @property
+    def x(self):
+        return self._x
+
+    @property
+    def y(self):
+        return self._y
+
+    @property
+    def width(self):
+        return self._width
+
+    @property
+    def height(self):
+        return self._height
+
+    @property
+    def cropped(self):
+        return self._width != self._full_width or self._height != self._full_height
