@@ -20,15 +20,10 @@ import cgi
 import re
 
 from . import warn, wrap, get_int_param, parse_int, open_file, SkoolParsingError
+from .skoolmacro import DELIMITERS
 
 DIRECTIVES = 'bcgituwz'
 
-HTML_MACRO_DELIMITERS = {
-    '(': ')',
-    '[': ']',
-    '{': '}'
-}
-
 TABLE_MARKER = '#TABLE'
 TABLE_END_MARKER = 'TABLE#'
 COLUMN_WRAP_MARKER = ':w'
@@ -283,7 +278,7 @@ class SkoolParser:
     def get_instruction_addr_str(self, address, asm_id=''):
         instruction = self.get_instruction(address, asm_id)
         if instruction:
-            return instruction.addr_str
+            return instruction.get_addr_str(self.base)
         if asm_id:
             return self.mode.get_addr_str(address)
 
@@ -404,12 +399,12 @@ class SkoolParser:
     def _html_escape(self, text):
         chunks = []
         while 1:
-            search = re.search('#HTML[^A-Z]', text)
+            search = re.search('(#HTML[^A-Z]|#FONT:.)', text)
             if not search:
                 break
             start, index = search.span()
             delim1 = search.group()[-1]
-            delim2 = HTML_MACRO_DELIMITERS.get(delim1, delim1)
+            delim2 = DELIMITERS.get(delim1, delim1)
             end = text.find(delim2, index) + 1
             if end < index:
                 end = len(text)
@@ -797,7 +792,7 @@ class Mode:
         address = parse_int(addr_str)
         if self.decimal:
             if address:
-                addr_str = str(address)
+                addr_str = '{:05d}'.format(address)
             if operation:
                 operation = self.convert(operation)
         elif self.hexadecimal:
@@ -953,6 +948,11 @@ class Instruction:
     def html_escape(self):
         self.operation = cgi.escape(self.operation)
 
+    def get_addr_str(self, base):
+        if (base is None and self.label[1].isdigit()) or base == BASE_10:
+            return re.sub('^0{1,4}', '', self.addr_str)
+        return self.addr_str
+
 class Reference:
     def __init__(self, entry, address, addr_str):
         self.entry = entry
@@ -1032,7 +1032,11 @@ class Data(SkoolEntry):
 
 class TableParser:
     def parse_text(self, text, index):
-        end = text.index(TABLE_END_MARKER, index) + len(TABLE_END_MARKER)
+        try:
+            end = text.index(TABLE_END_MARKER, index) + len(TABLE_END_MARKER)
+        except ValueError:
+            marker = text[text.rindex('#', 0, index):index]
+            raise SkoolParsingError("Missing table end marker: {}{}...".format(marker, text[index:index + 15]))
         return end, self.parse_table(text[index:end])
 
     def parse_table(self, table_def):
@@ -1154,7 +1158,10 @@ class Table:
                     prev_spans[col_index] = (prev_rowspan - 1, prev_colspan)
                 col_index += prev_colspan
 
-        self.num_cols = 1 + max([cell.col_index for cell in [row[-1] for row in self.rows]])
+        if self.rows:
+            self.num_cols = 1 + max([cell.col_index for cell in [row[-1] for row in self.rows]])
+        else:
+            self.num_cols = 0
         self._calculate_col_widths()
 
     def get_header_rows(self):
@@ -1228,7 +1235,10 @@ class Cell:
 
 class ListParser:
     def parse_text(self, text, index):
-        end = text.index(LIST_END_MARKER, index) + len(LIST_END_MARKER)
+        try:
+            end = text.index(LIST_END_MARKER, index) + len(LIST_END_MARKER)
+        except ValueError:
+            raise SkoolParsingError("No end marker: #LIST{}...".format(text[index:index + 15]))
         return end, self.parse_list(text[index:end])
 
     def parse_list(self, list_def):
