@@ -6,7 +6,7 @@ import unittest
 
 from skoolkittest import SkoolKitTestCase
 import skoolkit
-from skoolkit import skool2html, VERSION, SkoolKitError
+from skoolkit import normpath, skool2html, PACKAGE_DIR, VERSION, SkoolKitError
 from skoolkit.skoolhtml import HtmlWriter
 from skoolkit.skoolparser import CASE_UPPER, CASE_LOWER
 
@@ -231,6 +231,13 @@ class Skool2HtmlTest(SkoolKitTestCase):
         with self.assertRaisesRegexp(SkoolKitError, '{}: file not found'.format(skoolfile)):
             self.run_skool2html('-d {0} {1}'.format(self.odir, skoolfile))
 
+    def test_nonexistent_skool_file_named_in_ref_file(self):
+        skoolfile = 'pqr.skool'
+        ref = '[Config]\nSkoolFile={}'.format(skoolfile)
+        reffile = self.write_text_file(ref, suffix='.ref')
+        with self.assertRaisesRegexp(SkoolKitError, '{}: file not found'.format(skoolfile)):
+            self.run_skool2html('-d {} {}'.format(self.odir, reffile))
+
     def test_nonexistent_ref_file(self):
         reffile = 'zyx.ref'
         with self.assertRaisesRegexp(SkoolKitError, '{}: file not found'.format(reffile)):
@@ -280,6 +287,24 @@ class Skool2HtmlTest(SkoolKitTestCase):
         self.assertEqual(output[2], 'Using ref files: {0}, {1}'.format(reffile, reffile2))
         self.assertEqual(output[6], '  Writing disassembly files in {0}/{1}'.format(basename(prefix), code_path))
 
+    def test_config_in_secondary_ref_file(self):
+        self.mock(skool2html, 'write_disassembly', mock_write_disassembly)
+        skoolfile = self.write_text_file(suffix='.skool')
+        reffile1 = self.write_text_file(suffix='.ref')
+        game_dir = 'foo'
+        ref = '\n'.join((
+            '[Config]',
+            'SkoolFile={}'.format(skoolfile),
+            'GameDir={}'.format(game_dir)
+        ))
+        reffile2 = self.write_text_file(ref, '{}-2.ref'.format(reffile1[:-4]))
+        output, error = self.run_skool2html(reffile1)
+        self.assertEqual(error, '')
+        html_writer = write_disassembly_args[0]
+        self.assertEqual(html_writer.file_info.game_dir, game_dir)
+        skool_parser = html_writer.parser
+        self.assertEqual(skool_parser.skoolfile, skoolfile)
+
     def test_skool_from_stdin(self):
         self.write_stdin('; Routine\nc30000 RET')
         game_dir = 'program'
@@ -398,6 +423,27 @@ class Skool2HtmlTest(SkoolKitTestCase):
                 self.run_skool2html(reffile)
             self.assertEqual(cm.exception.args[0], 'Invalid colour spec: {}={}'.format(name, spec))
 
+    def test_image_writer_options(self):
+        self.mock(skool2html, 'write_disassembly', mock_write_disassembly)
+        exp_iw_options = (
+            ('DefaultFormat', 'gif'),
+            ('GIFCompression', 0),
+            ('GIFEnableAnimation', 0),
+            ('GIFTransparency', 1),
+            ('PNGAlpha', 1),
+            ('PNGCompressionLevel', 4),
+            ('PNGEnableAnimation', 0)
+        )
+        ref = '[ImageWriter]\n' + '\n'.join(['{}={}'.format(k, v) for k, v in exp_iw_options])
+        reffile = self.write_text_file(ref, suffix='.ref')
+        self.write_text_file(path='{}.skool'.format(reffile[:-4]))
+
+        output, error = self.run_skool2html(reffile)
+        self.assertEqual(error, '')
+        iw_options = write_disassembly_args[0].image_writer.options
+        for k, v in exp_iw_options:
+            self.assertEqual(iw_options[k], v)
+
     def test_font_is_copied(self):
         font_file = self.write_bin_file(suffix='.ttf')
         reffile = self.write_text_file("[Game]\nFont={}".format(font_file), suffix='.ref')
@@ -417,6 +463,150 @@ class Skool2HtmlTest(SkoolKitTestCase):
         self.assertTrue(os.path.isfile(os.path.join(game_dir, font_files[0])))
         self.assertTrue(os.path.isfile(os.path.join(game_dir, font_files[1])))
 
+    def test_custom_font_path(self):
+        font_file = self.write_bin_file(suffix='.ttf')
+        font_path = 'fonts'
+        ref = '\n'.join((
+            '[Game]',
+            'Font={}',
+            '[Paths]',
+            'FontPath={}'
+        )).format(font_file, font_path)
+        reffile = self.write_text_file(ref, suffix='.ref')
+        self.write_text_file(path='{}.skool'.format(reffile[:-4]))
+        output, error = self.run_skool2html('{} -d {} {}'.format(self._css_c(), self.odir, reffile))
+        self.assertEqual(error, '')
+        game_dir = os.path.join(self.odir, reffile[:-4])
+        self.assertTrue(os.path.isfile(os.path.join(game_dir, font_path, font_file)))
+
+    def test_custom_javascript_path(self):
+        js_file = self.write_text_file(suffix='.js')
+        js_path = 'javascript'
+        ref = '\n'.join((
+            '[Paths]',
+            'JavaScriptPath={}',
+            '[Page:CustomPage]',
+            'JavaScript={}',
+            'Path=page.html',
+            'PageContent=<b>Hello</b>'
+        )).format(js_path, js_file)
+        reffile = self.write_text_file(ref, suffix='.ref')
+        self.write_text_file(path='{}.skool'.format(reffile[:-4]))
+        output, error = self.run_skool2html('{} -d {} -w P {}'.format(self._css_c(), self.odir, reffile))
+        self.assertEqual(error, '')
+        game_dir = os.path.join(self.odir, reffile[:-4])
+        self.assertTrue(os.path.isfile(os.path.join(game_dir, js_path, js_file)))
+
+    def test_custom_style_sheet_path(self):
+        css_file = self.write_text_file(suffix='.css')
+        style_sheet_path = 'css'
+        ref = '\n'.join((
+            '[Game]',
+            'StyleSheet={}',
+            '[Paths]',
+            'StyleSheetPath={}'
+        )).format(css_file, style_sheet_path)
+        reffile = self.write_text_file(ref, suffix='.ref')
+        self.write_text_file(path='{}.skool'.format(reffile[:-4]))
+        output, error = self.run_skool2html('-d {} {}'.format(self.odir, reffile))
+        self.assertEqual(error, '')
+        game_dir = os.path.join(self.odir, reffile[:-4])
+        self.assertTrue(os.path.isfile(os.path.join(game_dir, style_sheet_path, css_file)))
+
+    def test_resources_are_copied(self):
+        resource1 = self.write_bin_file(suffix='.jpg')
+        resource2 = self.write_bin_file(suffix='.swf')
+        resource3 = self.write_bin_file(suffix='.pdf')
+        dest_dir = 'stuff'
+        ref = '\n'.join((
+            '[Resources]',
+            '{}=.'.format(resource1),
+            '{}={}'.format(resource2, dest_dir),
+            '{}={}'.format(resource3, dest_dir)
+        ))
+        reffile = self.write_text_file(ref, suffix='.ref')
+        self.write_text_file(path='{}.skool'.format(reffile[:-4]))
+        output, error = self.run_skool2html('{} -d {} {}'.format(self._css_c(), self.odir, reffile))
+        self.assertEqual(error, '')
+        game_dir = os.path.join(self.odir, reffile[:-4])
+        self.assertTrue(os.path.isfile(os.path.join(game_dir, resource1)))
+        self.assertTrue(os.path.isfile(os.path.join(game_dir, dest_dir, resource2)))
+        self.assertTrue(os.path.isfile(os.path.join(game_dir, dest_dir, resource3)))
+
+    def test_resource_not_found(self):
+        resource = 'nosuchfile.png'
+        reffile = self.write_text_file("[Resources]\n{}=foo".format(resource), suffix='.ref')
+        self.write_text_file(path='{}.skool'.format(reffile[:-4]))
+        with self.assertRaisesRegexp(SkoolKitError, 'Cannot copy resource "{}": file not found'.format(resource)):
+            self.run_skool2html('{} -d {} {}'.format(self._css_c(), self.odir, reffile))
+
+    def test_resource_is_not_copied_over_existing_file(self):
+        resource = self.write_bin_file(suffix='.jpg')
+        dest_dir = 'already-exists'
+        reffile = self.write_text_file("[Resources]\n{}={}".format(resource, dest_dir), suffix='.ref')
+        self.write_text_file(path='{}.skool'.format(reffile[:-4]))
+        path = os.path.join(self.odir, reffile[:-4], dest_dir)
+        self.write_text_file(path=path)
+        with self.assertRaisesRegexp(SkoolKitError, 'Cannot copy {0} to {1}: {1} is not a directory'.format(resource, dest_dir)):
+            self.run_skool2html('{} -d {} {}'.format(self._css_c(), self.odir, reffile))
+
+    def test_option_s(self):
+        exp_preamble = [
+            'skool2html.py searches the following directories for skool files, ref files,',
+            'CSS files, JavaScript files, font files, and files listed in the [Resources]',
+            'section of the ref file:',
+            ''
+        ]
+        preamble_len = len(exp_preamble)
+        exp_search_dirs = [
+            'The directory that contains the skool or ref file named on the command line',
+            'The current working directory',
+            os.path.join('.', 'resources'),
+            os.path.join(os.path.expanduser('~'), '.skoolkit'),
+            os.path.normpath('/usr/share/skoolkit'),
+            os.path.join(PACKAGE_DIR, 'resources')
+        ]
+        for option in ('-s', '--search-dirs'):
+            output, error = self.run_skool2html(option, catch_exit=0)
+            self.assertEqual(error, '')
+            self.assertEqual(output[:preamble_len], exp_preamble)
+            self.assertEqual(output[preamble_len:], ['- ' + d for d in exp_search_dirs])
+
+    def test_option_j(self):
+        css1_content = 'a { color: blue }'
+        css1 = self.write_text_file(css1_content, suffix='.css')
+        css2_content = 'td { color: red }'
+        css2 = self.write_text_file(css2_content, suffix='.css')
+        ref = '[Game]\nStyleSheet={};{}'.format(css1, css2)
+        reffile = self.write_text_file(ref, suffix='.ref')
+        self.write_text_file(path='{}.skool'.format(reffile[:-4]))
+        game_dir = normpath(self.odir, reffile[:-4])
+        single_css = 'style.css'
+        single_css_f = normpath(game_dir, single_css)
+        appending_msg = 'Appending {{}} to {}'.format(single_css_f)
+        for option in ('-j', '--join-css'):
+            output, error = self.run_skool2html('{} {} -d {} {}'.format(option, single_css, self.odir, reffile))
+            self.assertEqual(error, '')
+            self.assertTrue(appending_msg.format(css1) in output)
+            self.assertTrue(appending_msg.format(css2) in output)
+            self.assertTrue(os.path.isfile(single_css_f))
+            self.assertFalse(os.path.isfile(os.path.join(game_dir, css1)))
+            self.assertFalse(os.path.isfile(os.path.join(game_dir, css2)))
+            with open(single_css_f) as f:
+                css = f.read()
+            self.assertEqual(css, '\n'.join((css1_content, css2_content, '')))
+
+    def test_option_j_directory_exists(self):
+        single_css = 'game.css'
+        skoolfile = self.write_text_file(suffix='.skool')
+        game_dir = normpath(self.odir, skoolfile[:-6])
+        dest = normpath(game_dir, single_css)
+        self.make_directory(dest)
+        error_msg = "Cannot write CSS file '{}': {} already exists and is a directory".format(single_css, dest)
+        for option in ('-j', '--join-css'):
+            with self.assertRaisesRegexp(SkoolKitError, error_msg):
+                self.run_skool2html('{} {} {} -d {} {}'.format(option, single_css, self._css_c(), self.odir, skoolfile))
+
     def test_option_a(self):
         self.mock(skool2html, 'write_disassembly', mock_write_disassembly)
         self.mock(skool2html, 'SkoolParser', MockSkoolParser)
