@@ -658,6 +658,11 @@ class OptionsTest(SkoolKitTestCase):
         self.assertEqual(lines[1], '; @org=16384')
         self.assertEqual(lines[3][:16], 'c16384 LD BC,770')
 
+    def test_unrecognised_snapshot_format_is_treated_as_binary(self):
+        binfile = self.write_bin_file([1, 2, 3], suffix='.qux')
+        lines = self._write_skool(binfile, 3)
+        self.assertEqual(lines[3], 'c65533 LD BC,770     ;')
+
     def test_default_sft(self):
         # Test that the default skool file template is used if present
         binfile = self._write_bin([0])
@@ -673,6 +678,14 @@ class OptionsTest(SkoolKitTestCase):
         lines = self._write_skool('-c {0} {1}'.format(ctlfile, binfile), 3)
         self.assertEqual(lines[2], '; Control file')
 
+    def test_default_sft_for_unrecognised_snapshot_format(self):
+        self.mock(sna2skool, 'run', mock_run)
+        binfile = 'snapshot.foo'
+        sftfile = self.write_text_file(path='{}.sft'.format(binfile))
+        sna2skool.main((binfile,))
+        snafile, options = run_args
+        self.assertEqual(options.sftfile, sftfile)
+
     def test_default_ctl(self):
         # Test that the default control file is used if present
         binfile = self._write_bin([0])
@@ -689,6 +702,14 @@ class OptionsTest(SkoolKitTestCase):
         lines = self._write_skool('-T {0} {1}'.format(sftfile, binfile), 1)
         self.assertEqual(lines[0], sft[0])
 
+    def test_default_ctl_for_unrecognised_snapshot_format(self):
+        self.mock(sna2skool, 'run', mock_run)
+        binfile = 'input.bar'
+        ctlfile = self.write_text_file(path='{}.ctl'.format(binfile))
+        sna2skool.main((binfile,))
+        snafile, options = run_args
+        self.assertEqual(options.ctlfile, ctlfile)
+
     def test_terminal_unknown_block(self):
         data = (
             205, 255, 255, # 65531 CALL 65535
