@@ -1,8 +1,10 @@
 # -*- coding: utf-8 -*-
+import re
 import unittest
 
 from skoolkittest import SkoolKitTestCase
-from skoolkit.snaskool import Disassembly, SkoolWriter
+from skoolkit import SkoolKitError
+from skoolkit.snaskool import Disassembly, SkoolWriter, write_ctl
 from skoolkit.ctlparser import CtlParser
 
 DISASSEMBLY_SNAPSHOT = [0] * 65536
@@ -494,5 +496,124 @@ class SkoolWriterTest(SkoolKitTestCase):
         self.assertEqual(len(warnings), 1)
         self.assertEqual(warnings[0], 'WARNING: Code block at 65533 overlaps the following block at 65534')
 
+    def test_blank_multi_instruction_comment(self):
+        ctl = '\n'.join((
+            'c 65534',
+            '  65534,2 .'
+        ))
+        writer = self._get_writer([0] * 65536, ctl)
+        writer.write_skool(0, False)
+        skool = self.out.getvalue().split('\n')[:-1]
+        self.assertEqual(skool[-2], 'c65534 NOP           ; {')
+        self.assertEqual(skool[-1], ' 65535 NOP           ; }')
+
+    def test_instruction_comments_starting_with_a_dot(self):
+        comment1 = '...'
+        comment2 = '...and so it ends'
+        ctl = '\n'.join((
+            'c 65534',
+            '  65534,1 {}'.format(comment1),
+            '  65535,1 {}'.format(comment2)
+        ))
+        writer = self._get_writer([0] * 65536, ctl)
+        writer.write_skool(0, False)
+        skool = self.out.getvalue().split('\n')[:-1]
+        self.assertEqual(skool[-2], 'c65534 NOP           ; {}'.format(comment1))
+        self.assertEqual(skool[-1], ' 65535 NOP           ; {}'.format(comment2))
+
+    def test_multi_instruction_comments_starting_with_a_dot(self):
+        comment1 = '...'
+        comment2 = '...and so it ends'
+        ctl = '\n'.join((
+            'c 65532',
+            '  65532,2 .{}'.format(comment1),
+            '  65534,2 {}'.format(comment2)
+        ))
+        writer = self._get_writer([0] * 65536, ctl)
+        writer.write_skool(0, False)
+        skool = self.out.getvalue().split('\n')[:-1]
+        self.assertEqual(skool[-4], 'c65532 NOP           ; {{{}'.format(comment1))
+        self.assertEqual(skool[-3], ' 65533 NOP           ; }')
+        self.assertEqual(skool[-2], ' 65534 NOP           ; {{{}'.format(comment2))
+        self.assertEqual(skool[-1], ' 65535 NOP           ; }')
+
+    def test_decimal_addresses_below_10000(self):
+        ctl = '\n'.join((
+            'b 00000',
+            'i 00001',
+            'b 00003',
+            'i 00004',
+            'b 00023',
+            'i 00024',
+            'b 00573',
+            'i 00574',
+            'b 01876',
+            'i 01877',
+        ))
+        writer = self._get_writer([0] * 1877, ctl)
+        writer.write_skool(0, False)
+        skool = self.out.getvalue().split('\n')[:-1]
+        self.assertEqual(skool[3], 'b00000 DEFB 0')
+        self.assertEqual(skool[8], 'b00003 DEFB 0')
+        self.assertEqual(skool[13], 'b00023 DEFB 0')
+        self.assertEqual(skool[18], 'b00573 DEFB 0')
+        self.assertEqual(skool[23], 'b01876 DEFB 0')
+
+    def test_decimal_org_addresses_below_10000(self):
+        snapshot = [0] * 1235
+        for org in (0, 1, 12, 123, 1234):
+            ctl = 'b {:05d}\ni {:05d}'.format(org, org + 1)
+            writer = self._get_writer(snapshot, ctl)
+            self.clear_streams()
+            writer.write_skool(0, False)
+            skool = self.out.getvalue().split('\n')[:-1]
+            self.assertEqual(skool[1], '; @org={}'.format(org))
+
+    def test_no_table_end_marker(self):
+        ctl = '\n'.join((
+            'b 00000',
+            'D 00000 #TABLE { this table has no end marker }',
+            'i 00001'
+        ))
+        writer = self._get_writer([0], ctl)
+        with self.assertRaisesRegexp(SkoolKitError, re.escape("No end marker found: #TABLE { this table h...")):
+            writer.write_skool(0, False)
+
+    def test_no_udgtable_end_marker(self):
+        ctl = '\n'.join((
+            'b 00000',
+            'D 00000 #UDGTABLE { this table has no end marker }',
+            'i 00001'
+        ))
+        writer = self._get_writer([0], ctl)
+        with self.assertRaisesRegexp(SkoolKitError, re.escape("No end marker found: #UDGTABLE { this table h...")):
+            writer.write_skool(0, False)
+
+    def test_no_list_end_marker(self):
+        ctl = '\n'.join((
+            'b 00000',
+            'D 00000 #LIST { this list has no end marker }',
+            'i 00001'
+        ))
+        writer = self._get_writer([0], ctl)
+        with self.assertRaisesRegexp(SkoolKitError, re.escape("No end marker found: #LIST { this list ha...")):
+            writer.write_skool(0, False)
+
+class CtlWriterTest(SkoolKitTestCase):
+    def test_decimal_addresses_below_10000(self):
+        ctls = {0: 'b', 1: 'c', 22: 't', 333: 'w', 4444: 'z'}
+        exp_ctl = [
+            'b 00000',
+            'c 00001',
+            't 00022',
+            'w 00333',
+            'z 04444'
+        ]
+        ctlfile = self.write_bin_file()
+        write_ctl(ctlfile, ctls, False)
+        with open(ctlfile, 'r') as f:
+            lines = [line.rstrip() for line in f]
+        self.assertEqual(lines, exp_ctl)
+
 if __name__ == '__main__':
     unittest.main()
